r"""
An implementation of quantile regression neural networks based on the
`Keras <https://github.com/keras-team/keras>` deep learning package.
The implementation has been developed specifically for remote sensing
applications and provides a high level interface allowing for simple
training and evaluation of the QRNNs."""
from typhon.retrieval.qrnn.qrnn import QRNN
